import { normalizeTemplate, createSNode, isVoidElement, } from "./shared.js";
export const compileTemplate = (tpl) => {
    const roots = [];
    const children = normalizeTemplate(tpl.children);
    const exprMap = createExprMap(children);
    for (let i = 0; i < children.length; i++) {
        const child = children[i];
        switch (child.type) {
            case 0 /* INodeType.Element */:
                const sNode = createSNode(child, 0);
                roots.push(createTElement(exprMap, sNode));
                break;
            case 1 /* INodeType.Text */:
                roots.push(child.value);
                break;
            case 2 /* INodeType.Expr */:
                roots.push(exprMap.get(child.value));
                break;
        }
    }
    return {
        roots,
        exprs: Array.from(exprMap.keys()),
    };
};
const createExprMap = (children) => {
    const exprMap = new Map();
    for (let i = 0; i < children.length; i++) {
        _createExprMap(exprMap, children[i]);
    }
    return exprMap;
};
const _createExprMap = (exprMap, node) => {
    switch (node.type) {
        case 0 /* INodeType.Element */:
            const { tag, properties, children } = node;
            for (let i = 0; i < properties.length; i++) {
                const prop = properties[i];
                const value = prop.value;
                if (typeof value === "number" && isSerializableDynamicProp(tag, prop)) {
                    exprMap.set(value, exprMap.size);
                }
            }
            for (let i = 0; i < children.length; i++) {
                _createExprMap(exprMap, children[i]);
            }
            break;
        case 1 /* INodeType.Text */:
            break;
        case 2 /* INodeType.Expr */:
            exprMap.set(node.value, exprMap.size);
    }
};
const createTElement = (exprMap, sNode) => {
    const node = sNode.node;
    const { tag, properties } = node;
    let innerHTML;
    let suffix;
    let directives = null;
    let children = null;
    let props = null;
    let style = null;
    let flags = 0;
    let prefix = `<${tag}`;
    if (properties.length > 0) {
        let styleDynamic = null;
        let styleStatic = "";
        for (let i = 0; i < properties.length; i++) {
            const { key, value, type } = properties[i];
            if (type === 0 /* IPropertyType.Attribute */) {
                if (typeof value === "number") {
                    props = pushProp(props, ` ${key}`, 0, exprMap.get(value));
                }
                else if (value === true) {
                    prefix += ` ${key}`;
                }
                else {
                    if (key === "style") {
                        styleStatic += `${value};`;
                    }
                    else {
                        prefix += ` ${key}="${value}"`;
                    }
                }
            }
            else if (type === 1 /* IPropertyType.Value */ || type === 2 /* IPropertyType.DOMValue */) {
                if (key === "value") {
                    if (tag === "input") {
                        props = pushProp(props, ` value`, 1 /* TFlags.IgnoreEmptyString */, exprMap.get(value));
                    }
                    else if (tag === "textarea") {
                        flags |= 2 /* TFlags.EscapeInnerHTML */;
                        innerHTML = exprMap.get(value);
                    }
                }
                else if (key === "checked") {
                    if (tag === "input") {
                        props = pushProp(props, ` checked`, 0, exprMap.get(value));
                    }
                }
                else if (key === "selected") {
                    if (tag === "option") {
                        props = pushProp(props, ` selected`, 0, exprMap.get(value));
                    }
                }
                else if (key === "textContent") {
                    flags |= 2 /* TFlags.EscapeInnerHTML */;
                    innerHTML = exprMap.get(value);
                }
                else if (key === "innerHTML") {
                    innerHTML = exprMap.get(value);
                }
            }
            else if (type === 3 /* IPropertyType.Style */) {
                if (typeof value === "string") {
                    styleStatic += `${key}:${value};`;
                }
                else {
                    styleDynamic = pushProp(styleDynamic, `${key}:`, 0, exprMap.get(value));
                }
            }
            else if (type === 5 /* IPropertyType.Directive */) {
                if (key === 1 /* IDirectiveType.Server */) {
                    const j = exprMap.get(value);
                    if (directives === null) {
                        directives = [j];
                    }
                    else {
                        directives.push(j);
                    }
                }
            }
        }
        if (styleDynamic !== null) {
            style = {
                stat: styleStatic,
                dyn: styleDynamic,
            };
        }
        else if (styleStatic) {
            prefix += ` style="${styleStatic}"`;
        }
    }
    const sChildren = sNode.children;
    if (isVoidElement(tag)) {
        if (sChildren !== null) {
            throw new Error(`Invalid template, void element '${tag}' shouldn't have any children.`);
        }
        suffix = "";
    }
    else {
        if (innerHTML !== void 0) {
            children = innerHTML;
        }
        else if (sChildren !== null && sChildren.length > 0) {
            children = [];
            for (let i = 0; i < sChildren.length; i++) {
                const child = sChildren[i];
                switch (child.node.type) {
                    case 0 /* INodeType.Element */:
                        children.push(createTElement(exprMap, child));
                        break;
                    case 1 /* INodeType.Text */:
                        children.push(child.node.value);
                        break;
                    case 2 /* INodeType.Expr */:
                        if (child.flags & 4 /* SNodeFlags.HasNextDOMNode */) {
                            flags |= 1 /* TFlags.GenerateOffsets */;
                        }
                        children.push(exprMap.get(child.node.value));
                }
            }
        }
        suffix = `</${tag}>`;
    }
    return {
        flags,
        prefix,
        suffix,
        props,
        style,
        children,
        directives,
    };
};
const pushProp = (props, prefix, flags, i) => {
    if (props === null) {
        props = [];
    }
    props.push({ flags, prefix, i });
    return props;
};
const isSerializableDynamicProp = (tag, prop) => {
    const type = prop.type;
    if (type === 0 /* IPropertyType.Attribute */) {
        if (prop.key === "class" && prop.hoist === true) {
            return false;
        }
        return true;
    }
    if (type === 1 /* IPropertyType.Value */ || type === 2 /* IPropertyType.DOMValue */) {
        const key = prop.key;
        if (tag === "input") {
            if (key === "value" || key === "checked") {
                return true;
            }
        }
        else if (tag === "textarea") {
            if (key === "value") {
                return true;
            }
        }
        else if (tag === "option") {
            if (key === "selected") {
                return true;
            }
        }
        if (key === "textContent" || key === "innerHTML") {
            return true;
        }
        return false;
    }
    if (type === 3 /* IPropertyType.Style */) {
        return true;
    }
    if (type === 5 /* IPropertyType.Directive */) {
        return (prop.key === 1 /* IDirectiveType.Server */);
    }
    return false;
};
//# sourceMappingURL=server.js.map